/**
*  RTC4513 Clock Lib (by Andreas Boehme)
*
*  Bitte PORT/PIN Definitionen in rtc4513.h anpassen
*
**/
#include <inttypes.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include "rtc4513.h"

#define rtc_clk_high()   RTC_CLK  |=  _BV(RTC_CLK_PIN);
#define rtc_clk_low()    RTC_CLK  &= ~_BV(RTC_CLK_PIN);
#define rtc_data_high()  RTC_DATA |=  _BV(RTC_DATA_PIN);
#define rtc_data_low()   RTC_DATA  &= ~_BV(RTC_DATA_PIN);
#define rtc_en_high()    RTC_EN  |=  _BV(RTC_EN_PIN);
#define rtc_en_low()     RTC_EN  &= ~_BV(RTC_EN_PIN);

#define PIN(x) (*(&x - 2))
#define DDR(x) (*(&x - 1))

/*************************************************************************
Delay routines 
*************************************************************************/
static inline void _delayFourCycles(unsigned int __count)
{
    if ( __count == 0 )    
        __asm__ __volatile__( "rjmp 1f\n 1:" );    // 2 cycles
    else
        __asm__ __volatile__ (
    	    "1: sbiw %0,1" "\n\t"                  
    	    "brne 1b"                              // 4 cycles/loop
    	    : "=w" (__count)
    	    : "0" (__count)
    	   );
}

#define delay(us)  _delayFourCycles( ( ( 1*(XTAL/4000) )*us)/1000 )

/*
  toggle Data bit on PORT
*/
static void output_bit(unsigned char bit) {
  DDR(RTC_DATA) |= _BV(RTC_DATA_PIN);       
  if (bit == 1) {
    rtc_data_high();
  } else {
    rtc_data_low();
  }
  
}

/*
  Send one Nibble (4 Bits) to RTC
*/
static void send_nibble(unsigned char data) {
    unsigned char i;
    for(i=0; i<4; i++) {
        output_bit(((data>>i)&0x01));
        delay(3);
        rtc_clk_high();
        delay(3);
        rtc_clk_low();
        delay(3);
    }
}


/*
  Receive one Nibble (4 Bits) from RTC
*/

static unsigned char receive_nibble(void) {
    unsigned char data = 0;
    unsigned char i;
    DDR(RTC_DATA) &= ~_BV(RTC_DATA_PIN);
    for(i=0; i<4; i++) {
        rtc_clk_high();
        delay(3);
        if (PIN(RTC_DATA) & _BV(RTC_DATA_PIN))
          data += (1 << i);
        rtc_clk_low();
        delay(3);
    }
    return data;
}

/*
  send one byte to RTC
*/
static void send_byte(unsigned char data) {
    send_nibble(data);
    send_nibble(data>>4);
}

/*
  write count bytes to the given address
*/
static void rtc_write(unsigned char addr, unsigned char *data, unsigned char count) {
    unsigned char i;
    rtc_en_high();
    send_nibble(0x03);
    send_nibble(addr);
    for(i=0; i < count; i++) {
        send_byte(data[i]);
    }
    rtc_en_low();
}

/*
  read count bytes from rtc to data buffer
*/
static void rtc_read(unsigned char addr, unsigned char *data, unsigned char count) {
    unsigned char i;
    rtc_en_high();
    send_nibble(0x0c);
    send_nibble(addr);
    for(i=0; i < count; i++) {
        data[i] = receive_nibble();
        data[i] |= (receive_nibble()<<4);
    }
   rtc_en_low();
}

/*
  Start / Stop Clock
*/
static void rtc_start(unsigned char start) {
    if(start) {
        rtc_en_high();
        send_nibble(0x03);
        send_nibble(CFREG);
        send_nibble(0x04);
        rtc_en_low();
        delay(3);
        rtc_en_high();
        send_nibble(0x03);
        send_nibble(CDREG);
        send_nibble(0x02);
        rtc_en_low();
    } else {
        rtc_en_high();
        send_nibble(0x03);
        send_nibble(CFREG);
        send_nibble(0x07);
        rtc_en_low();
    }
}

/*
  Convert a byte to BCD (binary coded digit)
*/
static unsigned char byte2bcd(unsigned char data) {
    return ((data/10)<<4) + (data%10);
}

/*
  and back
*/
static unsigned char bcd2byte(unsigned char bcd) {
    return (bcd>>4)*10 + (bcd&0x0f);
}

/*
  set the RTC - TIME
*/
void settime(unsigned char hrs, unsigned char min, unsigned char sec) {
    unsigned char data[3];
    data[0] = byte2bcd(sec);
    data[1] = byte2bcd(min);
    data[2] = byte2bcd(hrs);
    rtc_start(0);
    rtc_write(SEC1REG, data, 3);
    rtc_start(1);
}

/*
  set the RTC - DATE
*/
void setdate(unsigned char day, unsigned char month, unsigned char year) {
    unsigned char data[3];
    data[0] = byte2bcd(day);
    data[1] = byte2bcd(month);
    data[2] = byte2bcd(year);
    rtc_start(0);
    rtc_write(DAY1REG, data, 3);
    rtc_start(1);
}


/*
  set Day Of Week
*/
void setdow(unsigned char dow) {
    rtc_start(0);
    rtc_en_high();
    send_nibble(0x03);
    send_nibble(DOWREG);
    send_nibble(dow);
    rtc_en_low();
    rtc_start(1);
}

/*
  Read time from RTC to buffer (decoded to byte)
  time[0] -> hour, time[1] -> minute, time[2] -> second
*/
void gettime(unsigned char* time) {
    static unsigned char data[3];
        do {
            rtc_read(SEC1REG, data, 3);
        } while(data[1]&0x80 || data[2]&0x80);

    time[2] = bcd2byte(data[0]&0x7f);
    time[1] = bcd2byte(data[1]);
    time[0] = bcd2byte(data[2]&0x3f);
}

/*
  if the clock and data lines are used for other serial communication,
  it is good to have a "known state" before reading / writing to RTC
*/
void rtc_set_bus(void) {
    DDR(RTC_DATA) |= _BV(RTC_DATA_PIN);
    DDR(RTC_CLK) |= _BV(RTC_CLK_PIN);           
    DDR(RTC_EN) |= _BV(RTC_EN_PIN);
    rtc_en_low();    
    rtc_clk_low();
    rtc_data_low();
}


/* 
  RTC_START and BUS INIT, CALL BEFORE USE!
*/
void rtc_init(void) {
    rtc_set_bus();
    delay(100);
    rtc_start(1);
}

